import numpy as np
import pytest

from relife.lifetime_model import Exponential
from relife.policy import AgeReplacementPolicy, OneCycleAgeReplacementPolicy


class TestOneCycleAgeReplacementPolicy:
    # ignore runtime warning in optimization
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    def test_asymptotic_expected_equivalent_annual_cost(self, distribution, cf, cp, discounting_rate):
        if isinstance(distribution, Exponential):
            pytest.skip("Exponential distribution won't work with this cf, cp (not tested in v1.0.0 too)")
        policy = OneCycleAgeReplacementPolicy(distribution, cf, cp, discounting_rate=discounting_rate)
        try:
            policy.optimize()
        except RuntimeError:
            pytest.skip("Optimization failed, EEAC may be too flat")
        qa = policy.asymptotic_expected_equivalent_annual_cost()
        assert qa.shape == np.broadcast_shapes(cf.shape, cp.shape)  # () or (m,)

    # ignore runtime warning in optimization
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    def test_expected_equivalent_annual_cost(self, distribution, cf, cp, discounting_rate):
        if isinstance(distribution, Exponential):
            pytest.skip("Exponential distribution won't work with this cf, cp (not tested in v1.0.0 too)")
        policy = OneCycleAgeReplacementPolicy(distribution, cf, cp, discounting_rate=discounting_rate)
        try:
            policy.optimize()
        except RuntimeError:
            pytest.skip("Optimization failed, EEAC may be too flat")

        qa = policy.asymptotic_expected_equivalent_annual_cost()  # () or (m,)
        timeline, q = policy.expected_equivalent_annual_cost(400, nb_steps=2000)

        assert timeline.shape == (2000,)
        assert q.shape == qa.shape + timeline.shape  # (m, 2000) or (2000,)

    def test_optimal_replacement_age(self, distribution, cf, cp, discounting_rate):
        if isinstance(distribution, Exponential):
            pytest.skip("Exponential distribution won't work with this cf, cp (not tested in v1.0.0 too)")
        eps = 1e-2
        policy = OneCycleAgeReplacementPolicy(distribution, cf, cp, discounting_rate=discounting_rate).optimize()
        policy1 = OneCycleAgeReplacementPolicy(
            distribution,
            cf,
            cp,
            discounting_rate=discounting_rate,
            period_before_discounting=0.1,
            ar=policy.ar + eps,
        )
        policy0 = OneCycleAgeReplacementPolicy(
            distribution,
            cf,
            cp,
            discounting_rate=discounting_rate,
            period_before_discounting=0.1,
            ar=policy.ar - eps,
        )
        assert np.all(
            policy1.asymptotic_expected_equivalent_annual_cost() > policy.asymptotic_expected_equivalent_annual_cost()
        ) and np.all(
            policy0.asymptotic_expected_equivalent_annual_cost() > policy.asymptotic_expected_equivalent_annual_cost()
        )


class TestAgeReplacementPolicy:
    # ignore runtime warning in optimization
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    def test_asymptotic_expected_equivalent_annual_cost(self, distribution, cf, cp, discounting_rate):
        if isinstance(distribution, Exponential):
            pytest.skip("Exponential distribution won't work with this cf, cp (not tested in v1.0.0 too)")
        policy = AgeReplacementPolicy(distribution, cf, cp, discounting_rate=discounting_rate)
        try:
            policy.optimize()
        except RuntimeError:
            pytest.skip("Optimization failed, EEAC may be too flat")
        qa = policy.asymptotic_expected_equivalent_annual_cost()  # () or (m,)
        assert qa.shape == np.broadcast_shapes(cf.shape, cp.shape)  # () or (m,)

    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    def test_expected_equivalent_annual_cost(self, distribution, cf, cp, discounting_rate):
        if isinstance(distribution, Exponential):
            pytest.skip("Exponential distribution won't work with this cf, cp (not tested in v1.0.0 too)")
        policy = AgeReplacementPolicy(distribution, cf, cp, discounting_rate=discounting_rate)
        try:
            policy.optimize()
        except RuntimeError:
            pytest.skip("Optimization failed, EEAC may be too flat")
        qa = policy.asymptotic_expected_equivalent_annual_cost()  # () or (m,)
        timeline, q = policy.expected_equivalent_annual_cost(400, nb_steps=2000)

        assert timeline.shape == (2000,)
        assert q.shape == qa.shape + timeline.shape  # (m, 2000) or (2000,)
        assert q[..., -1].flatten() == pytest.approx(qa.flatten(), rel=1e-1)

    def test_optimal_replacement_age(self, distribution, cf, cp, discounting_rate):
        if isinstance(distribution, Exponential):
            pytest.skip("Exponential distribution won't work with this cf, cp (not tested in v1.0.0 too)")
        eps = 1e-2
        policy = AgeReplacementPolicy(distribution, cf, cp, discounting_rate=discounting_rate).optimize()
        ar = policy.ar

        policy1 = AgeReplacementPolicy(distribution, cf, cp, discounting_rate=discounting_rate, ar=ar + eps)
        policy0 = AgeReplacementPolicy(distribution, cf, cp, discounting_rate=discounting_rate, ar=ar - eps)

        assert np.all(
            policy1.asymptotic_expected_equivalent_annual_cost() > policy.asymptotic_expected_equivalent_annual_cost()
        ) and np.all(
            policy0.asymptotic_expected_equivalent_annual_cost() > policy.asymptotic_expected_equivalent_annual_cost()
        )



from typing import (
    Callable,
    Literal,
    Optional,
    Protocol,
    TypeAlias,
    TypeVarTuple,
    overload,
)

import numpy as np
from numpy.typing import NDArray

_Xs = TypeVarTuple("_Xs")
_X: TypeAlias = float | NDArray[np.float64]
_Y: TypeAlias = np.float64 | NDArray[np.float64]
_B: TypeAlias = np.bool_ | NDArray[np.bool_]

# M = TypeVar("M", bound=int)
# N = TypeVar("N", bound=int)
# _Shape = TypeVarTuple("_Shape")
# _Array : TypeAlias = np.ndarray[tuple[*_Shape], np.dtype[np.float64]]
#
# _Scalar : TypeVarTuple = np.float64
# _1DArray : TypeAlias = _Array[M]
# _2DArray : TypeAlias = _Array[M, N]


class _ParametricLifetimeModel(Protocol[*_Xs]):

    def hf(self, time: _X, *args: *_Xs) -> _Y: ...
    def chf(self, time: _X, *args: *_Xs) -> _Y: ...
    def sf(self, time: _X, *args: *_Xs) -> _Y: ...
    def pdf(self, time: _X, *args: *_Xs) -> _Y: ...
    def mrl(self, time: _X, *args: *_Xs) -> _Y: ...
    def moment(self, n: int, *args: *_Xs) -> _Y: ...
    def mean(self, *args: *_Xs) -> _Y: ...
    def var(self, *args: *_Xs) -> _Y: ...
    def isf(self, probability: _X, *args: *_Xs) -> _Y: ...
    def ichf(self, cumulative_hazard_rate: _X, *args: *_Xs) -> _Y: ...
    def cdf(self, time: _X, *args: *_Xs) -> _Y: ...
    def ppf(self, probability: _X, *args: *_Xs) -> _Y: ...
    def median(self, *args: *_Xs) -> _Y: ...
    def ls_integrate(
        self,
        func: Callable[[_X], _Y],
        a: _X,
        b: _X,
        *args: *_Xs,
        deg: int = 10,
    ) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B]: ...
    @overload
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _Y]: ...
    @overload
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B, _Y]: ...
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> _Y | tuple[_Y, _Y] | tuple[_Y, _B] | tuple[_Y, _B, _Y]: ...



import numpy as np
import pytest
from pytest import approx

from relife.economic import RunToFailureReward
from relife.lifetime_model import EquilibriumDistribution
from relife.stochastic_process import RenewalProcess, RenewalRewardProcess


class TestDistribution:
    def test_renewal_density(self, distribution):
        renewal_process = RenewalProcess(distribution, first_lifetime_model=EquilibriumDistribution(distribution))
        timeline, renewal_density = renewal_process.renewal_density(100, 200)
        assert timeline.shape == (200,)
        assert renewal_density.shape == (200,)
        assert renewal_density[..., -1:] == approx(1 / distribution.mean(), rel=1e-4)

    def test_expected_total_reward(self, distribution):
        reward = RunToFailureReward(cf=1.0)
        renewal_reward_process = RenewalRewardProcess(distribution, reward)
        timeline_m, m = renewal_reward_process.renewal_function(100.0, 200)
        assert timeline_m.shape == m.shape == (200,)
        timeline_z, z = renewal_reward_process.expected_total_reward(100, 200)
        assert timeline_z.shape == z.shape == (200,)
        assert m == approx(z, rel=1e-4)

    def test_renewal_reward_process_vec(self, distribution):
        cf0 = 1
        n = 3
        cf = cf0 / n

        rrp0 = RenewalRewardProcess(
            distribution,
            RunToFailureReward(cf0),
            discounting_rate=0.04,
        )
        rrp = RenewalRewardProcess(
            distribution,
            RunToFailureReward(np.full((n, 1), cf)),
            discounting_rate=0.04,
        )

        timeline_z, z = rrp.expected_total_reward(100, 200)  # (3, nb_steps)
        assert timeline_z.shape == (200,)
        assert z.shape == (3, 200)
        timeline_z0, z0 = rrp0.expected_total_reward(100, 200)  # (nb_steps,)
        assert timeline_z0.shape == (200,)
        assert z0.shape == (200,)
        assert z.shape == (n, 200)
        assert z0 == approx(z.sum(axis=0), rel=1e-4)

    @pytest.mark.skip(reason="was not tested in v1 and we will work on it later")
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    def test_sample_lifetime_data(self, distribution):
        expected_params = distribution.params.copy()
        q1 = distribution.ppf(0.25)
        q3 = distribution.ppf(0.75)
        success = 0
        n = 100
        renewal_process = RenewalProcess(distribution)
        for i in range(n):
            lifetime_data = renewal_process.generate_lifetime_data(10 * q3, t0=0, size=10000)
            try:  #  for gamma and loglogistic essentially (convergence errors may occcur)
                distribution.fit(**lifetime_data)
            except RuntimeError:
                continue
            ic = distribution.fitting_results.IC
            # params found are within params IC
            print("expected params :", expected_params)
            print("IC95 :", ic)
            if np.all(expected_params.reshape(-1, 1) >= ic[:, [0]]) and np.all(
                expected_params.reshape(-1, 1) <= ic[:, [1]]
            ):
                success += 1
        print(success / n)
        assert success >= 0.95 * n


class TestAgeReplacementDistribution:
    def test_renewal_density(self, frozen_ar_distribution):
        renewal_process = RenewalProcess(
            frozen_ar_distribution, first_lifetime_model=EquilibriumDistribution(frozen_ar_distribution)
        )
        timeline, renewal_density = renewal_process.renewal_density(100, 200)
        assert timeline.shape == (200,)
        assert renewal_density.shape == (200,)
        assert renewal_density[..., -1:] == approx(1 / frozen_ar_distribution.mean(), rel=1e-4)

    def test_expected_total_reward(self, frozen_ar_distribution):
        reward = RunToFailureReward(cf=1.0)
        renewal_reward_process = RenewalRewardProcess(frozen_ar_distribution, reward)
        timeline_m, m = renewal_reward_process.renewal_function(100.0, 200)
        assert timeline_m.shape == (200,)
        assert m.shape == (200,)
        timeline_z, z = renewal_reward_process.expected_total_reward(100, 200)
        assert timeline_z.shape == (200,)
        assert z.shape == (200,)
        assert m == approx(z, rel=1e-4)

    def test_renewal_reward_process_vec(self, frozen_ar_distribution):
        cf0 = 1
        n = 3
        cf = cf0 / n

        rrp0 = RenewalRewardProcess(
            frozen_ar_distribution,
            RunToFailureReward(cf0),
            discounting_rate=0.04,
        )
        rrp = RenewalRewardProcess(
            frozen_ar_distribution,
            RunToFailureReward(np.full((n, 1), cf)),
            discounting_rate=0.04,
        )
        timeline_z, z = rrp.expected_total_reward(100, 200)  # (3, nb_steps)
        assert timeline_z.shape == (200,)
        assert z.shape == (3, 200)
        timeline_z0, z0 = rrp0.expected_total_reward(100, 200)  # (nb_steps,)
        assert timeline_z0.shape == (200,)
        assert z0.shape == (200,)
        assert z.shape == (n, 200)
        assert z0 == approx(z.sum(axis=0), rel=1e-4)


class TestRegression:
    def test_renewal_density(self, frozen_regression):
        renewal_process = RenewalProcess(
            frozen_regression, first_lifetime_model=EquilibriumDistribution(frozen_regression)
        )
        timeline, renewal_density = renewal_process.renewal_density(100, 200)
        assert timeline.shape == (200,)
        assert renewal_density.shape == (3, 200)
        assert renewal_density[..., -1:] == approx(1 / frozen_regression.mean(), rel=1e-4)

    def test_expected_total_reward(self, frozen_regression):
        reward = RunToFailureReward(cf=1.0)
        renewal_reward_process = RenewalRewardProcess(frozen_regression, reward)
        timeline_m, m = renewal_reward_process.renewal_function(100.0, 200)
        assert timeline_m.shape == (200,)
        assert m.shape == (3, 200)
        timeline_z, z = renewal_reward_process.expected_total_reward(100, 200)
        assert timeline_z.shape == (200,)
        assert z.shape == (3, 200)
        assert m == approx(z, rel=1e-4)

    def test_renewal_reward_process_vec(self, frozen_regression):
        cf0 = 1
        n = 3
        cf = cf0 / n

        rrp0 = RenewalRewardProcess(
            frozen_regression,
            RunToFailureReward(cf0),
            discounting_rate=0.04,
        )
        rrp = RenewalRewardProcess(
            frozen_regression,
            RunToFailureReward(np.full((n, 1), cf)),
            discounting_rate=0.04,
        )
        timeline_z, z = rrp.expected_total_reward(100, 200)  # (3, nb_steps)

        assert timeline_z.shape == (200,)
        assert z.shape == (3, 200)

        timeline_z0, z0 = rrp0.expected_total_reward(100, 200)  # (3, nb_steps,)

        assert timeline_z0.shape == (200,)
        assert z0.shape == (n, 200)
        assert z.shape == (n, 200)
        assert z0 == approx(n * z, rel=1e-4)


class TestAgeReplacementRegression:
    def test_renewal_density(self, frozen_ar_regression):
        renewal_process = RenewalProcess(
            frozen_ar_regression, first_lifetime_model=EquilibriumDistribution(frozen_ar_regression)
        )
        timeline, renewal_density = renewal_process.renewal_density(100, 200)
        assert timeline.shape == (200,)
        assert renewal_density.shape == (3, 200)
        assert renewal_density[..., -1:] == approx(1 / frozen_ar_regression.mean(), rel=1e-4)

    def test_expected_total_reward(self, frozen_ar_regression):
        reward = RunToFailureReward(cf=1.0)
        renewal_reward_process = RenewalRewardProcess(frozen_ar_regression, reward)
        timeline_m, m = renewal_reward_process.renewal_function(100.0, 200)
        assert timeline_m.shape == (200,)
        assert m.shape == (3, 200)
        timeline_z, z = renewal_reward_process.expected_total_reward(100, 200)
        assert timeline_z.shape == (200,)
        assert z.shape == (3, 200)
        assert m == approx(z, rel=1e-4)

    def test_renewal_reward_process_vec(self, frozen_ar_regression):
        cf0 = 1
        n = 3
        cf = cf0 / n

        rrp0 = RenewalRewardProcess(
            frozen_ar_regression,
            RunToFailureReward(cf0),
            discounting_rate=0.04,
        )
        rrp = RenewalRewardProcess(
            frozen_ar_regression,
            RunToFailureReward(np.full((n, 1), cf)),
            discounting_rate=0.04,
        )
        timeline_z, z = rrp.expected_total_reward(100, 200)  # (3, nb_steps)
        assert timeline_z.shape == (200,)
        assert z.shape == (3, 200)
        timeline_z0, z0 = rrp0.expected_total_reward(100, 200)  # (3, nb_steps,)
        assert timeline_z0.shape == (200,)
        assert z0.shape == (
            n,
            200,
        )
        assert z.shape == (n, 200)
        assert z0 == approx(n * z, rel=1e-4)



from __future__ import annotations

from typing import Callable, Optional

import numpy as np
from numpy.typing import NDArray

from relife.economic import Discounting, ExponentialDiscounting
from relife.lifetime_model import (
    FrozenAgeReplacementModel,
    FrozenLeftTruncatedModel,
)
from relife.lifetime_model.distribution import LifetimeDistribution
from relife.lifetime_model.regression import FrozenLifetimeRegression


def renewal_equation_solver(
    timeline: NDArray[np.float64],
    lifetime_model: LifetimeDistribution | FrozenLifetimeRegression | FrozenAgeReplacementModel,
    evaluated_func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
    discounting: Optional[Discounting] = None,
) -> NDArray[np.float64]:

    # timeline : (nb_steps,) or (m, nb_steps)
    tm = 0.5 * (timeline[..., 1:] + timeline[..., :-1])  # (nb_steps - 1,) or (m, nb_steps - 1)
    f = lifetime_model.cdf(timeline)  # (nb_steps,) or (m, nb_steps)
    fm = lifetime_model.cdf(tm)  # (nb_steps - 1,) or (m, nb_steps - 1)
    y = evaluated_func(timeline)  # (nb_steps,) or (m, nb_steps)

    if y.shape != f.shape:
        raise ValueError("Invalid shape between model and evaluated_func")

    if discounting is not None:
        d = discounting.factor(timeline)  # (nb_steps,) or (m, nb_steps)
    else:
        d = np.ones_like(f)
    z = np.empty(y.shape)
    u = d * np.insert(f[..., 1:] - fm, 0, 1, axis=-1)
    v = d[..., :-1] * np.insert(np.diff(fm), 0, 1, axis=-1)
    q0 = 1 / (1 - d[..., 0] * fm[..., 0])
    z[..., 0] = y[..., 0]
    z[..., 1] = q0 * (y[..., 1] + z[..., 0] * u[..., 1])
    for n in range(2, f.shape[-1]):
        z[..., n] = q0 * (y[..., n] + z[..., 0] * u[..., n] + np.sum(z[..., 1:n][..., ::-1] * v[..., 1:n], axis=-1))
    return z


def delayed_renewal_equation_solver(
    timeline: NDArray[np.float64],
    z: NDArray[np.float64],
    first_lifetime_model: (
        LifetimeDistribution | FrozenLifetimeRegression | FrozenAgeReplacementModel | FrozenLeftTruncatedModel
    ),
    evaluated_func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
    discounting: Optional[ExponentialDiscounting] = None,
) -> NDArray[np.float64]:
    # timeline : (nb_steps,) or (m, nb_steps)
    tm = 0.5 * (timeline[..., 1:] + timeline[..., :-1])  # (nb_steps - 1,) or (m, nb_steps - 1)
    f1 = first_lifetime_model.cdf(timeline)  # (nb_steps,) or (m, nb_steps)
    f1m = first_lifetime_model.cdf(tm)  # (nb_steps - 1,) or (m, nb_steps - 1)
    y1 = evaluated_func(timeline)  # (nb_steps,) or (m, nb_steps - 1)
    if discounting is not None:
        d = discounting.factor(timeline)  # (nb_steps,) or (m, nb_steps - 1)
    else:
        d = np.ones_like(f1)
    z1 = np.empty(y1.shape)
    u1 = d * np.insert(f1[..., 1:] - f1m, 0, 1, axis=-1)
    v1 = d[..., :-1] * np.insert(np.diff(f1m), 0, 1, axis=-1)
    z1[..., 0] = y1[..., 0]
    z1[..., 1] = y1[..., 1] + z[..., 0] * u1[..., 1] + z[..., 1] * d[..., 0] * f1m[..., 0]
    for n in range(2, f1.shape[-1]):
        z1[..., n] = (
            y1[..., n]
            + z[..., 0] * u1[..., n]
            + z[..., n] * d[..., 0] * f1m[..., 0]
            + np.sum(z[..., 1:n][..., ::-1] * v1[..., 1:n], axis=-1)
        )
    return z1



from itertools import chain
from typing import Self

import numpy as np


class ParametricModel:
    """
    Base class to create a parametric_model core.

    Any parametric_model core must inherit from `ParametricModel`.
    """

    def __init__(self, **kwparams):
        self._params = Parameters(**kwparams)
        self._nested_models = {}

    @property
    def params(self):
        """
        Parameters values.

        Returns
        -------
        ndarray
            Parameters values of the core

        Notes
        -----
        If parameter values are not set, they are encoded as `np.nan` value.

        Parameters can be by manually setting`params` through its setter, fitting the core if `fit` exists or
        by specifying all parameters values when the core object is initialized.
        """
        return np.array(self._params.get_values())

    @params.setter
    def params(self, new_params):
        if not isinstance(new_params, np.ndarray):
            raise ValueError(
                f"Incorrect params values. It must be contained in a 1d array. Got type {type(new_params)}"
            )
        if new_params.ndim > 1:
            raise ValueError(f"Expected params values to be 1d array. Got {new_params.ndim} ndim")
        self._params.set_values(tuple(v.item() for v in new_params))

    @property
    def params_names(self):
        """
        Parameters names.

        Returns
        -------
        list of str
            Parameters names

        Notes
        -----
        Parameters values can be requested (a.k.a. get) by their name at instance level.
        """
        return self._params.get_names()

    @property
    def nb_params(self):
        """
        Number of parameters.

        Returns
        -------
        int
            Number of parameters.

        """
        return self._params.size

    def _set_nested_model(self, name, model):
        """Compose with new ``ParametricModel`` instance(s).

        This method acts like function. When you set a nested model instance, the followings happen :
        - each new parameters are added to the current ``Parameters`` instance
        - each new `ParametricModel` instance is accessible as a standard attribute

        Parameters
        ----------
        name : str
            model name
        model : ParametricModel
            model instance
        """
        self._nested_models[name] = model
        self._params.set_leaf(f"{name}.params", getattr(model, "_params"))

    def __getattr__(self, name):
        class_name = type(self).__name__
        if name in self.__dict__:
            return self.__dict__[name]
        if name in super().__getattribute__("_params").get_names():
            return super().__getattribute__("_params").get_param_value(name)
        if name in super().__getattribute__("_nested_models"):
            return super().__getattribute__("_nested_models").get(name)
        raise AttributeError(f"{class_name} has no attribute named {name}")

    def __setattr__(self, name, value):
        if name in ("_params", "_nested_models"):
            super().__setattr__(name, value)
        elif name in self._params.get_names():
            self._params.set_param_value(name, value)
        elif isinstance(value, ParametricModel):
            self._set_nested_model(name, value)
        else:  # just set a new attribute
            super().__setattr__(name, value)


# MAYBE, custom array container can replace it : https://numpy.org/doc/stable/user/basics.dispatch.html#writing-custom-array-containers
class Parameters:
    """
    Dict-like tree structured parameters.

    Every ``ParametricModel`` are composed of a ``Parameters`` instance.
    """

    def __init__(self, **kwargs):
        self._parent = None
        self._leaves = {}
        self._nodemapping = {}
        self._values = ()
        self._names = ()
        if bool(kwargs):
            self.set_node(kwargs)

    def get_names(self):
        return self._names

    def get_values(self):
        return self._values

    def get_leaf(self, name: str):
        try:
            return self._leaves[name]
        except KeyError:
            raise ValueError(f"Parameters object does not have leaf parameters called {name} in its scope")

    def get_param_value(self, name):
        try:
            return self._nodemapping[name]
        except KeyError:
            raise ValueError(f"Parameters object does not have parameter name called {name} in its scope")

    def set_param_value(self, name, value):
        if name not in self._nodemapping:
            raise ValueError(f"Parameters object does not have parameter name called {name} in its scope")
        self._nodemapping[name] = value
        self._update_names_and_values()

    @property
    def size(self):
        return len(self._values)

    def set_node(self, mapping):
        """
        set node dict
        """
        self._nodemapping = {k: v if v is not None else np.nan for k, v in mapping.items()}
        self._update_names_and_values()  # update _names and _values

    def set_leaf(self, leaf_name, leaf):
        """
        set a leaf or new leaf
        """
        if leaf_name not in self._leaves:
            leaf._parent = self
        self._leaves[leaf_name] = leaf
        self._update_names_and_values()  # update _names and _values

    def set_values(self, values):
        """set values of all tree"""
        if len(values) != self.size:
            raise ValueError(f"Expected {self.size} values but got {len(values)}")
        pos = len(self._nodemapping)
        self._nodemapping.update(zip(self._nodemapping.keys(), (np.nan if v is None else v for v in values[:pos])))
        self._values = tuple((np.nan if v is None else v for v in values))
        for leaf in self._leaves.values():
            leaf.set_values(values[pos : pos + leaf.size])
            pos += leaf.size
        if self._parent is not None:
            self._parent._update_names_and_values()

    def _update_names_and_values(self):
        """update names and values of current and parent nodes"""

        def items_walk(parameters: Self):
            yield tuple(parameters._nodemapping.items())
            for leaf in parameters._leaves.values():
                yield tuple(chain.from_iterable(items_walk(leaf)))

        generator = chain.from_iterable(items_walk(self))
        try:
            self._names, self._values = zip(*generator)
        except StopIteration:
            pass
        # recursively update parent
        if self._parent is not None:
            self._parent._update_names_and_values()


def _get_nb_assets(*args):
    if not bool(args):
        return 1
    args_2d = tuple((np.atleast_2d(arys) for arys in args))

    try:
        broadcast_shape = np.broadcast_shapes(*(ary.shape for ary in args_2d))
        nb_assets = broadcast_shape[0]
    except ValueError:
        raise ValueError("Given args have incompatible shapes")
    return nb_assets


class FrozenParametricModel(ParametricModel):
    def __init__(self, model, *args):
        super().__init__()
        if np.any(np.isnan(model.params)):
            raise ValueError("You try to freeze a model with unsetted parameters. Set params first")
        self.nb_assets = _get_nb_assets(*args)
        self.args = args
        self.unfrozen_model = model

    def unfreeze(self):
        return self.unfrozen_model


def is_frozen(model):
    return isinstance(model, FrozenParametricModel)


def is_lifetime_model(model):
    from relife.lifetime_model import (
        NonParametricLifetimeModel,
        ParametricLifetimeModel,
    )

    return isinstance(model, (ParametricLifetimeModel, NonParametricLifetimeModel))


def is_stochastic_process(model):
    from relife.stochastic_process import StochasticProcess

    return isinstance(model, StochasticProcess)


# see sklearn/base.py : return unfitted ParametricModel
# def clone(model: ParametricModel) -> ParametricModel: ...



from ._base import (
    FittableParametricLifetimeModel,
    FrozenParametricLifetimeModel,
    NonParametricLifetimeModel,
    ParametricLifetimeModel,
)
from .conditional_model import (
    AgeReplacementModel,
    FrozenAgeReplacementModel,
    FrozenLeftTruncatedModel,
    LeftTruncatedModel,
)
from .distribution import (
    EquilibriumDistribution,
    Exponential,
    Gamma,
    Gompertz,
    LifetimeDistribution,
    LogLogistic,
    MinimumDistribution,
    Weibull,
)
from .non_parametric import ECDF, KaplanMeier, NelsonAalen, Turnbull
from .regression import (
    AcceleratedFailureTime,
    CovarEffect,
    FrozenLifetimeRegression,
    LifetimeRegression,
    ProportionalHazard,
)



from typing import Callable, Literal, Optional, overload

import numpy as np
from numpy.typing import NDArray
from typing_extensions import override

from relife._typing import _B, _X, _Y, _ParametricLifetimeModel, _Xs

from ._base import FrozenParametricLifetimeModel, ParametricLifetimeModel

def reshape_ar_or_a0(name: str, value: _X) -> NDArray[np.float64]: ...

class AgeReplacementModel(ParametricLifetimeModel[*tuple[_X, *_Xs]]):
    baseline: _ParametricLifetimeModel[*_Xs]
    def __init__(self, baseline: _ParametricLifetimeModel[*_Xs]) -> None: ...
    def sf(self, time: _X, ar: _X, *args: *_Xs) -> _Y: ...
    def hf(self, time: _X, ar: _X, *args: *_Xs) -> _Y: ...
    def chf(self, time: _X, ar: _X, *args: *_Xs) -> _Y: ...
    def cdf(self, time: _X, ar: _X, *args: *_Xs) -> _Y: ...
    def pdf(self, time: _X, ar: _X, *args: *_Xs) -> _Y: ...
    @override
    def isf(self, probability: _X, ar: _X, *args: *_Xs) -> _Y: ...
    @override
    def ichf(self, cumulative_hazard_rate: _X, ar: _X, *args: *_Xs) -> _Y: ...
    @override
    def ppf(self, probability: _X, ar: _X, *args: *_Xs) -> _Y: ...
    @override
    def median(self, ar: _X, *args: *_Xs) -> _Y: ...
    @override
    def mrl(self, time: _X, ar: _X, *args: *_Xs) -> _Y: ...
    @override
    def moment(self, n: int, ar: _X, *args: *_Xs) -> _Y: ...
    @override
    def mean(self, ar: _X, *args: *_Xs) -> _Y: ...
    @override
    def var(self, ar: _X, *args: *_Xs) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        ar: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        ar: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B]: ...
    @overload
    def rvs(
        self,
        size: int,
        ar: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _Y]: ...
    @overload
    def rvs(
        self,
        size: int,
        ar: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B, _Y]: ...
    def rvs(
        self,
        size: int,
        ar: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> _Y | tuple[_Y, _Y] | tuple[_Y, _B] | tuple[_Y, _B, _Y]: ...
    @override
    def ls_integrate(
        self,
        func: Callable[[_X], _Y],
        a: _X,
        b: _X,
        ar: _X,
        *args: *_Xs,
        deg: int = 10,
    ) -> _Y: ...
    def freeze(self, ar: _X, *args: *_Xs) -> FrozenAgeReplacementModel[*_Xs]: ...

class LeftTruncatedModel(ParametricLifetimeModel[*tuple[_X, *_Xs]]):
    baseline: _ParametricLifetimeModel[*_Xs]
    def __init__(self, baseline: _ParametricLifetimeModel[*_Xs]) -> None: ...
    def sf(self, time: _X, a0: _X, *args: *_Xs) -> _Y: ...
    def hf(self, time: _X, a0: _X, *args: *_Xs) -> _Y: ...
    def chf(self, time: _X, a0: _X, *args: *_Xs) -> _Y: ...
    def cdf(self, time: _X, ar: _X, *args: *_Xs) -> _Y: ...
    def pdf(self, time: _X, a0: _X, *args: *_Xs) -> _Y: ...
    @override
    def isf(self, probability: _X, a0: _X, *args: *_Xs) -> _Y: ...
    @override
    def ichf(self, cumulative_hazard_rate: _X, a0: _X, *args: *_Xs) -> _Y: ...
    @override
    def ppf(self, probability: _X, a0: _X, *args: *_Xs) -> _Y: ...
    @override
    def median(self, a0: _X, *args: *_Xs) -> _Y: ...
    @override
    def mrl(self, time: _X, a0: _X, *args: *_Xs) -> _Y: ...
    @override
    def moment(self, n: int, a0: _X, *args: *_Xs) -> NDArray[np.float64]: ...
    @override
    def mean(self, a0: _X, *args: *_Xs) -> NDArray[np.float64]: ...
    @override
    def var(self, a0: _X, *args: *_Xs) -> NDArray[np.float64]: ...
    @overload
    def rvs(
        self,
        size: int,
        a0: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        a0: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B]: ...
    @overload
    def rvs(
        self,
        size: int,
        a0: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _Y]: ...
    @overload
    def rvs(
        self,
        size: int,
        a0: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B, _Y]: ...
    def rvs(
        self,
        size: int,
        a0: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> _Y | tuple[_Y, _Y] | tuple[_Y, _B] | tuple[_Y, _B, _Y]: ...
    @override
    def ls_integrate(
        self,
        func: Callable[[_X], _Y],
        a: _X,
        b: _X,
        a0: _X,
        *args: *_Xs,
        deg: int = 10,
    ) -> _Y: ...
    def freeze(self, a0: _X, *args: *_Xs) -> FrozenLeftTruncatedModel[*_Xs]: ...

class FrozenAgeReplacementModel(FrozenParametricLifetimeModel[*tuple[_X, *_Xs]]):
    unfrozen_model: AgeReplacementModel[*_Xs]
    args: tuple[_X, *_Xs]
    @override
    def __init__(self, model: AgeReplacementModel[*_Xs], ar: _X, *args: *_Xs) -> None: ...
    @override
    def unfreeze(self) -> AgeReplacementModel[*_Xs]: ...
    @property
    def ar(self) -> _X: ...
    # noinspection PyUnresolvedReferences
    @ar.setter
    def ar(self, value: _X) -> None: ...

class FrozenLeftTruncatedModel(FrozenParametricLifetimeModel[*tuple[_X, *_Xs]]):
    unfrozen_model: LeftTruncatedModel[*_Xs]
    args: tuple[_X, *_Xs]
    @override
    def __init__(self, model: LeftTruncatedModel[*_Xs], a0: _X, *args: *_Xs) -> None: ...
    @override
    def unfreeze(self) -> LeftTruncatedModel[*_Xs]: ...
    @property
    def a0(self) -> _X: ...
    # noinspection PyUnresolvedReferences
    @a0.setter
    def a0(self, value: _X) -> None: ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Iterator, Optional

import numpy as np
from numpy.lib import recfunctions as rfn
from numpy.typing import NDArray
from typing_extensions import override

from relife.economic import ExponentialDiscounting, Reward
from relife.lifetime_model import (
    Exponential,
    FrozenAgeReplacementModel,
    FrozenLeftTruncatedModel,
)
from relife.lifetime_model.distribution import LifetimeDistribution
from relife.lifetime_model.regression import FrozenLifetimeRegression

if TYPE_CHECKING:
    from relife.stochastic_process import RenewalRewardProcess


class CountDataIterator(Iterator[NDArray[np.void]], ABC):

    def __init__(
        self,
        size: int,
        tf: float,
        t0: float = 0.0,
        nb_assets: Optional[int] = None,
        seed: Optional[int] = None,
    ):
        if nb_assets is None:
            np_size = (1, size)
        else:
            np_size = (nb_assets, size)
        self.t0, self.tf = t0, tf
        self.seed = seed
        # broadcasting of the size must be done before instanciation
        self.timeline = np.zeros(np_size, dtype=np.float64)

        self._crossed_t0_counter = np.zeros(np_size, dtype=np.uint32)
        self._crossed_tf_counter = np.zeros(np_size, dtype=np.uint32)
        self._observed_event = np.zeros(np_size, dtype=np.bool_)
        self.replacement_cycle = 0

        self.mask = np.zeros(
            np_size,
            dtype=np.dtype(
                [
                    ("observed_event", np.bool_),
                    ("just_crossed_t0", np.bool_),
                    ("just_crossed_tf", np.bool_),
                    ("crossed_tf", np.bool_),
                ]
            ),
        )

    def update_mask(self) -> None:
        self._crossed_t0_counter[self.timeline > self.t0] += 1
        self._crossed_tf_counter[self.timeline > self.tf] += 1
        self.mask["just_crossed_t0"] = self._crossed_t0_counter == 1
        self.mask["just_crossed_tf"] = self._crossed_tf_counter == 1
        self.mask["crossed_tf"] = self._crossed_tf_counter >= 1
        self.mask["observed_event"] = np.atleast_2d(
            np.logical_and(self._crossed_t0_counter >= 1, self._crossed_tf_counter <= 1)
        )

    def get_base_structarray(self) -> NDArray[np.void]:
        asset_id, sample_id = np.where(self.mask["observed_event"])

        struct_array = np.zeros(
            sample_id.size,
            dtype=np.dtype(
                [
                    ("sample_id", np.uint32),  #  unsigned 32bit integer
                    ("asset_id", np.uint32),  #  unsigned 32bit integer
                    ("timeline", np.float64),  #  64bit float
                ]
            ),
        )
        struct_array["timeline"] = self.timeline[self.mask["observed_event"]]
        struct_array["sample_id"] = sample_id.astype(np.uint32)
        struct_array["asset_id"] = asset_id.astype(np.uint32)
        return struct_array

    @abstractmethod
    def step(self) -> NDArray[np.void]: ...

    def __next__(self) -> NDArray[np.void]:
        if not np.all(self.mask["crossed_tf"]):
            struct_arr = self.step()
            while struct_arr.size == 0:  # skip cycles while arrays are empty (if t0 != 0.)
                struct_arr = self.step()
                if np.all(self.mask["crossed_tf"]):
                    raise StopIteration
            return struct_arr
        raise StopIteration


class RenewalProcessIterator(CountDataIterator):

    def __init__(
        self,
        process,
        size: int,
        tf: float,
        t0: float = 0.0,
        nb_assets: Optional[int] = None,
        seed: Optional[int] = None,
    ):
        super().__init__(size, tf, t0=t0, nb_assets=nb_assets, seed=seed)
        self.process = process

    @property
    def lifetime_model(self):
        if self.replacement_cycle == 0 and self.process.first_lifetime_model is not None:
            return self.process.first_lifetime_model
        return self.process.lifetime_model

    def step(self) -> NDArray[np.void]:
        # time is not residual age if return_entry is True (see LeftTruncatedModel)
        # time may be ar if model is AgeReplacementModel, then event is False
        # for first_cycle, model_entry may not be zero
        time, event, model_entry = self.lifetime_model.rvs(
            size=self.timeline.shape[1],
            nb_assets=self.timeline.shape[0],
            return_event=True,
            return_entry=True,
            seed=self.seed,
        )
        if self.replacement_cycle == 0 and np.any(model_entry != 0):
            # fixed age process
            self.timeline.fill(self.t0)
        residual_time = time - model_entry

        self.timeline += residual_time
        self.update_mask()

        # compute tf right censorings
        time = np.where(self.mask["just_crossed_tf"], time - (self.timeline - self.tf), time)
        self.timeline[self.mask["just_crossed_tf"]] = self.tf
        event[self.mask["just_crossed_tf"]] = False

        # compute t0 left truncations
        if self.replacement_cycle == 0:
            entry = np.where(self.mask["just_crossed_t0"], model_entry, 0.0)
        else:
            entry = np.where(self.mask["just_crossed_t0"], time - (self.timeline - self.t0), 0.0)

        # update seed to avoid having the same lifetime rvs
        if self.seed is not None:
            self.seed += 1
        self.replacement_cycle += 1

        struct_arr = rfn.append_fields(  #  works on structured_array too
            self.get_base_structarray(),
            ("time", "event", "entry"),
            (
                time[self.mask["observed_event"]],
                event[self.mask["observed_event"]],
                entry[self.mask["observed_event"]],
            ),
            (np.float64, np.bool_, np.float64),
            usemask=False,
            asrecarray=False,
        )
        return struct_arr


class RenewalRewardProcessIterator(RenewalProcessIterator):
    reward: Reward
    discounting: ExponentialDiscounting

    def __init__(
        self,
        process: RenewalRewardProcess[
            LifetimeDistribution | FrozenLifetimeRegression | FrozenAgeReplacementModel | FrozenLeftTruncatedModel,
            Reward,
        ],
        nb_samples: int,
        tf: float,
        t0: float = 0.0,
        nb_assets: Optional[int] = None,
        seed: Optional[int] = None,
    ):
        super().__init__(process, nb_samples, tf, t0=t0, nb_assets=nb_assets, seed=seed)
        self.reward = self.process.reward
        self.discounting = self.process.discounting

    @override
    def __next__(self) -> NDArray[np.void]:
        struct_array = super().__next__()
        # may be type hint error in rfn.append_fields overload
        return rfn.append_fields(
            struct_array,
            "reward",
            self.reward.sample(struct_array["time"]) * self.discounting.factor(struct_array["timeline"]),
            np.float64,
            usemask=False,
            asrecarray=False,
        )  # type: ignore


class NonHomogeneousPoissonProcessIterator(CountDataIterator):

    def __init__(
        self,
        process,
        nb_samples: int,
        tf: float,
        t0: float = 0.0,
        nb_assets: Optional[int] = None,
        ar: Optional[float | NDArray[np.float64]] = None,
        seed: Optional[int] = None,
    ):
        super().__init__(nb_samples, tf, t0=t0, nb_assets=nb_assets, seed=seed)
        self.process = process
        self.hpp_timeline = np.zeros_like(self.timeline, dtype=np.float64)
        self.failure_instant = np.zeros_like(self.timeline, dtype=np.float64)
        self.age = np.zeros_like(self.timeline, dtype=np.float64)
        self.entry = np.zeros_like(self.timeline, dtype=np.float64)
        self.is_new_asset = np.zeros_like(
            self.timeline, dtype=np.bool_
        )  # boolean flags indicating which assets are replaced
        self.renewals_ids = np.zeros_like(self.timeline, dtype=np.uint32)
        self.ar = (
            np.asarray(ar, dtype=np.float64).reshape(-1, 1)
            if ar is not None
            else np.full_like(self.timeline, np.inf, dtype=np.float64)
        )
        self.exponential_dist = Exponential(1.0)

    def step(self) -> NDArray[np.void]:

        # reset those who are replaced
        self.age[self.is_new_asset] = 0.0  # asset is replaced (0 aged asset)
        # self._assets_ids[self.is_new_asset] += 1 # asset is replaced (new asset id)
        self.hpp_timeline[self.is_new_asset] = 0.0
        self.failure_instant[self.is_new_asset] = 0.0
        self.entry[self.is_new_asset] = 0.0
        self.renewals_ids[self.is_new_asset] += 1
        self.is_new_asset.fill(False)  # reset to False

        # generate new values
        self.hpp_timeline += self.exponential_dist.rvs(
            self.timeline.shape[1], nb_assets=self.timeline.shape[0], seed=self.seed
        )
        failure_instant = self.process.unfrozen_model.lifetime_model.ichf(
            self.hpp_timeline, *getattr(self.process, "args", ())
        )
        duration = failure_instant - self.failure_instant  # t_i+1 - t_i
        self.failure_instant = failure_instant.copy()  # update t_i <- t_i+1
        self.timeline += duration
        self.age += duration

        # ar update (before because it changes timeline, thus start and stop conditions)
        self.timeline = np.where(
            self.age >= self.ar,
            self.timeline - (self.age - self.ar),
            self.timeline,
        )  # substract time after ar
        self.age = np.where(self.age >= self.ar, self.ar, self.age)
        self.is_new_asset[np.logical_and(self.age >= self.ar, ~self.mask["just_crossed_tf"])] = True
        event = np.where(np.logical_and(self.age >= self.ar, ~self.mask["just_crossed_tf"]), False, True)

        self.update_mask()

        # compute tf right censorings
        self.age = np.where(self.mask["just_crossed_tf"], self.age - (self.timeline - self.tf), self.age)
        self.timeline[self.mask["just_crossed_tf"]] = self.tf
        event[self.mask["just_crossed_tf"]] = False

        # compute t0 left truncations
        self.entry = np.where(self.mask["just_crossed_t0"], self.age - (self.timeline - self.t0), self.entry)

        entry = self.entry.copy()  # returned entry
        self.entry = np.where(event, self.age, self.entry)  # keep previous ages as entry for next iteration

        # update seed to avoid having the same rvs result
        if self.seed is not None:
            self.seed += 1

        struct_arr = rfn.append_fields(  #  works on structured_array too
            self.get_base_structarray(),
            ("age", "event", "entry"),
            (
                self.age[self.mask["observed_event"]],
                event[self.mask["observed_event"]],
                entry[self.mask["observed_event"]],
            ),
            (np.float64, np.bool_, np.float64),
            usemask=False,
            asrecarray=False,
        )
        return struct_arr
